from sklearn.model_selection import train_test_split, GridSearchCV, KFold
from sklearn.preprocessing import OneHotEncoder
from catboost import CatBoostClassifier
from ..metrics import Metrics
import optuna.visualization as vis
from sklearn.metrics import accuracy_score, roc_auc_score, log_loss
import pandas as pd
import numpy as np
import optuna
optuna.logging.disable_default_handler()

class Cat(Metrics):
    def __init__(self):
        self.metric = Metrics()
        self.model = None
        self.parameters = None

    def create(self, X, y, params=None):
        if params == None:
            cat = CatBoostClassifier()
            cat.fit(X,y)
            self.model = cat
        else:
            cat = CatBoostClassifier(**params)
            cat.fit(X,y)
            self.model = cat
            self.parameters = params

    def create_grid(self,X, y, params=None,cv=3, ram_limit=8):
        ram_limit *= 1024*1024*1024
        params_columns = ["depth","l2_leaf_reg","random_strength",
                          "thread_count","bootstrap_type","scale_pos_weight",
                          "user_ram_limit","verbose"]
        
        params_basic = {
            'depth': [3,7,10],
            'l2_leaf_reg': [1,10],
            'random_strength': [0.1,0.9],
            'thread_count': [-1],
            'bootstrap_type': ['Bayesian', 'Bernoulli', 'MVS'],
            'scale_pos_weight': [1, 3],
            'used_ram_limit': [ram_limit],
            'num_boost_round': 100,
            'verbose': [0],
            'random_state': [42]}
        
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        cat = CatBoostClassifier()
        grid_search = GridSearchCV(cat, params, cv=cv)
        grid_search.fit(X, y)
        best_params = grid_search.best_params_
        self.model = grid_search.best_estimator_
        self.parameters = best_params

    def create_optuna(self,X,y,params=None,n_trials=5,ram_limit=5, show_plot=False, show_features=False):
        ram_limit *= 1024*1024*1024
        params_columns = ["depth","l2_leaf_reg","random_strength",
                          "thread_count","bootstrap_type","scale_pos_weight",
                          "used_ram_limit","verbose","random_state",'num_boost_round']
        
        params_basic = {
            'depth': [1,8],
            'l2_leaf_reg': [0,10],
            'random_strength': [0.01,0.9],
            'thread_count': -1,
            'bootstrap_type': ['Bayesian', 'Bernoulli', 'MVS'],
            'scale_pos_weight': [0, 5],
            'num_boost_round': 200,
            'used_ram_limit': ram_limit,
            'verbose': 0,
            'random_state': 42}
        
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.2,random_state=42)
        def objective_binary(trial):
            param = {
                "depth": trial.suggest_int("depth", params["depth"][0], params["depth"][1]),
                "l2_leaf_reg": trial.suggest_float("l2_leaf_reg", params["l2_leaf_reg"][0], params["l2_leaf_reg"][1]),
                "random_strength": trial.suggest_float("random_strength", params["random_strength"][0], params["random_strength"][1]),
                "scale_pos_weight": trial.suggest_float("scale_pos_weight", params["scale_pos_weight"][0], params["scale_pos_weight"][1]),
                "bootstrap_type": trial.suggest_categorical("bootstrap_type", params["bootstrap_type"]),
                "used_ram_limit": trial.suggest_int("used_ram_limit", params["used_ram_limit"], params["used_ram_limit"]),
                "verbose": trial.suggest_int("verbose", params["verbose"], params["verbose"]),
                "thread_count": trial.suggest_int("thread_count", params["thread_count"], params["thread_count"]),
                "random_state": trial.suggest_int("random_state", params['random_state'], params['random_state']),
                'num_boost_round': trial.suggest_int("num_boost_round", params['num_boost_round'], params['num_boost_round'])
            }
            cat = CatBoostClassifier(**param)
            cat.fit(X_train,y_train)
            preds = cat.predict(X_test)
            loss = log_loss(y_test, preds)
            return loss

        def objective_multi(trial):
          param = {
                "depth": trial.suggest_int("depth", params["depth"][0], params["depth"][1]),
                "l2_leaf_reg": trial.suggest_float("l2_leaf_reg", params["l2_leaf_reg"][0], params["l2_leaf_reg"][1]),
                "random_strength": trial.suggest_float("random_strength", params["random_strength"][0], params["random_strength"][1]),
                "scale_pos_weight": trial.suggest_float("scale_pos_weight", params["scale_pos_weight"][0], params["scale_pos_weight"][1]),
                "bootstrap_type": trial.suggest_categorical("bootstrap_type", params["bootstrap_type"]),
                "used_ram_limit": trial.suggest_int("used_ram_limit", params["used_ram_limit"], params["used_ram_limit"]),
                "verbose": trial.suggest_int("verbose", params["verbose"], params["verbose"]),
                "thread_count": trial.suggest_int("thread_count", params["thread_count"], params["thread_count"]),
                "random_state": trial.suggest_int("random_state", params['random_state'], params['random_state']),
                'num_boost_round': trial.suggest_int("num_boost_round", params['num_boost_round'], params['num_boost_round'])
            }
          cat = CatBoostClassifier(**param,loss_function='MultiClass')
          cat.fit(X_train,y_train)
          preds = cat.predict_proba(X_test)
          loss = log_loss(y_test, preds)
          return loss


        study = optuna.create_study(direction="minimize", pruner=optuna.pruners.MedianPruner())
        if max(y) > 1:
          study.optimize(objective_multi, n_trials=n_trials)
          best_params = study.best_params
          cat = CatBoostClassifier(**best_params,loss_function='MultiClass')
        else:
          study.optimize(objective_binary, n_trials=n_trials)
          best_params = study.best_params
          cat = CatBoostClassifier(**best_params)

        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()

        cat.fit(X, y)
        self.model = cat
        self.parameters = best_params


    def score(self,X,y):
        preds = np.round(self.model.predict(X))
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None,classes=1):
        X, y = pd.DataFrame(X), pd.DataFrame(y)
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(shape=(df_test.shape[0],classes+1))
        roc = []
        n=0
        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y.iloc[valid_index]
            self.create(X_train,y_train,params=params)
            print("here", predictions.shape)
            print("preds", self.get().predict_proba(df_test).shape)
            print(self.get().predict_proba(df_test))
            predictions += self.get().predict_proba(df_test)/n_splits

            val_pred = self.get().predict_proba(X_test)
            if classes > 2:
              roc.append(roc_auc_score(y_test,val_pred,multi_class='ovr'))
            else:
              roc.append(roc_auc_score(y_test,val_pred))
            print(f"{i} Fold scored: {roc[i]}")

        print(f"Mean roc score {np.mean(roc)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters

    def set_targets(self,y):
      enc = OneHotEncoder(handle_unknown='ignore')
      y = enc.fit_transform(y.values.reshape(-1,1))
      targets = {key: value for key,value in enumerate(enc.categories_[0])}
      self.targets = targets
      return pd.DataFrame(y.toarray())

    def label_target(self,y):
      return [self.targets[np.argmax(i)] for i in y]

    def get_target(self):
      return self.targets

    def label2number(self, y):
      target_number = {}
      values = []
      for k,v in self.targets.items():
        target_number[v] = k
      for i in y:
        values.append(target_number[i])
      return values
    
    def save(self, model_path="cat.cbm"):
        self.model.save_model(model_path)

    def load(self, model_path):
        self.model.load_model(model_path)


from transformers import BertTokenizer, GPT2Tokenizer, RobertaTokenizer, DistilBertTokenizer, AlbertTokenizer, ElectraTokenizer
import numpy as np
import tensorflow as tf

class Text2Number():
    def __init__(self):
        self.token = None

    def tokenize_bert(self, data, seq_len=128, model="bert-base-cased"):
        '''
            Computationally expensive due to their large size, and the tokenization process can be slower
        '''
        tokenizer = BertTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_gpt2(self, data, seq_len=128, model="gpt2-medium"):
        '''
            Unidirectional
        '''
        tokenizer = GPT2Tokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_robert(self, data, seq_len=128, model="roberta-base"):
        '''
            Requiring more computational resources for training and inference.
        '''
        tokenizer = RobertaTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_distilBERT(self, data, seq_len=128, model="distilbert-base-uncased"):
        '''
            Scenarios where memory and computational resources are limited
        '''
        tokenizer = DistilBertTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_Albert(self, data, seq_len=128, model="albert-base-v2"):
        '''
            Fewer parameters, making them more memory-efficient and faster during training and inference
        '''
        tokenizer = AlbertTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def tokenize_Electra(self, data, seq_len=128, model="google/electra-base-discriminator"):
        '''
            Rely on a discriminator-generator setup, which means they may not capture certain aspects of language semantics
        '''
        tokenizer = ElectraTokenizer.from_pretrained(model)
        token = tokenizer(data.tolist(),
                          max_length=seq_len,
                          truncation=True,
                          padding='max_length',
                          add_special_tokens=True,
                          return_tensors="tf")
        self.token = token

    def create_dataset(self, labels, train=0.8, test=0.1,  batch_size=32):
        '''
            Creates the train and validation datasets
        '''
        labels = np.array(labels.tolist()).reshape(-1,1)
        size = int(labels.shape[0] / batch_size * train)
        size_test = int(labels.shape[0] / batch_size * test)

        dataset = tf.data.Dataset.from_tensor_slices((self.token['input_ids'], self.token['attention_mask'], labels))
        dataset = dataset.map(self.map_func_train)
        dataset = dataset.shuffle(10000).batch(batch_size,drop_remainder=True)

        train_dataset = dataset.take(size)
        remaining_dataset = dataset.skip(size)
        val_dataset = remaining_dataset.take(size_test)
        test_dataset = remaining_dataset.skip(size_test)
        return train_dataset, val_dataset, test_dataset
    
    def create_Xy(self,labels, batch_size=32):
        labels = np.array(labels.tolist()).reshape(-1,1)
        dataset = tf.data.Dataset.from_tensor_slices((self.token['input_ids'], self.token['attention_mask']))
        dataset = dataset.map(self.map_func_data)
        dataset = dataset.shuffle(10000).batch(batch_size,drop_remainder=True)
        return dataset, labels
    
    def map_func_train(self, input_ids, masks, labels):
        return {'input_ids':input_ids,
                'attention_mask':masks}, labels
    
    def map_func_data(self, input_ids, masks):
        return {'input_ids':input_ids, 'attention_mask':masks}
    
    def get(self):
        return self.token
    


from .cat import *
from .extra_trees import *
from .gradient_boosting import *
from .hist_gradient import *
from .kneighbours import *
from .light_lgb import *
from .linear_svc import *
from .logistic_regression import *
from .random_forest import *
from .sgd_classifier import *
from .svc import *
from .xgb import *


from typing import List
import seaborn as sns
from scipy.stats import norm
from sklearn.preprocessing import OneHotEncoder, MinMaxScaler
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import RFE
from sklearn.svm import SVR

class ModelData:
  def __init__(self, df):
    self.df_model = df

  def return_dataframe(self):
    return self.df_model

  def head(self):
    return self.df_model.head()

  def columns(self):
    return self.df_model.columns

  def unique(self, column):
    return self.df_model[column].unique()

  def bar(self, column):
    for col in column:
      self.df_model[column].value_counts().sort_values().plot(kind = "barh")
      plt.show()

  def sample(self,frac):
    return self.df_model.sample(frac=frac)
  
  def info(self):
    return self.df_model.info()
  
  def describe(self):
    return self.df_model.describe()

  def null(self):
    return self.df_model.isnull().sum()

  def find_null_column(self) -> List:
    columns = self.df_model.columns[self.df_model.isnull().any()]
    return columns

  def dropnull(self):
    self.df_model.dropna(inplace=True)

  def column_type(self, columns):
    for col in columns:
      print(f"{col}: {self.df_model[col].dtypes} \n")

  def dtypes_type(self, type):
    return self.df_model.select_dtypes(include=[type]).columns
  
  def rename(self, replace, replace_with):
    self.df_model = self.df_model.rename(columns={replace: replace_with})
  
  def drop(self, column):
    self.df_model = self.df_model.drop(column, axis=1)

  def fix_columns(self, columns:List[str] , method: str):
    if method == "mean":
      for column in columns:
        self.df_model[column].fillna(self.df_model[column].mean(), inplace=True)
    elif method == "zero":
      for column in columns:
        self.df_model[column].fillna(0, inplace=True)

  def fix_column_str(self, column):
    self.df_model[column] = self.df_model[column].replace('', "0")

  def fix_columns_values(self,column, threshold):
    for col, thr in zip(column, threshold):
      self.df_model[col] = self.df_model[col].apply(lambda x: self.df_model[col].mean() if x < thr else x)

  def count_column_condition(self, column1, column2, condition):
    '''
      Count specific value in another column (column2) when we group by column2 
      given condition that aplies to column2

      Inputs:
        -column1: Column to group by
        -column2: Column in which we trying to count elements
        -condition: given condition to column2

      Returns:
        - List of columns and count of elements for them
    '''
    return self.df_model.groupby(column1)[column2].apply(lambda x: (x == condition).sum())

  def delete_columns(self, columns: List[str]):
    for i in columns:
      self.df_model = self.df_model.drop(i, axis=1)

  def dtypes(self):
    return self.df_model.dtypes

  def corr(self):
    plt.figure(figsize=(16,16))
    sns.heatmap(data=self.df_model.corr(), annot=True, fmt=".2f")

  def normal_distribution(self, columns):
    for i in columns:
      col = self.df_model[i]
      mean, std_dev = norm.fit(col)
      plt.figure(figsize=(10,10))
      plt.hist(col, bins=20, density=True, alpha=0.6, color='g')
      xmin, xmax = plt.xlim()
      x = np.linspace(xmin, xmax, 100)
      p = norm.pdf(x, mean, std_dev)
      plt.plot(x, p, 'k', linewidth=2)
      title = f"Normal distribution of {i}: mean = {mean:.2f}, std dev = {std_dev:.2f}"
      plt.title(title)
      plt.show()

  def normal_distribution_values(self, column, column_search):
    values = self.df_model[column].unique()
    for val in values:
      data_new = self.df_model[self.df_model[column] == val]
      col = data_new[column_search]
      mean, std_dev = norm.fit(col)
      plt.figure(figsize=(10,10))
      plt.hist(col, bins=20, density=True, alpha=0.6, color='g')
      xmin, xmax = plt.xlim()
      x = np.linspace(xmin, xmax, 100)
      p = norm.pdf(x, mean, std_dev)
      plt.plot(x, p, 'k', linewidth=2)
      title = f"Normal distribution of {column_search} by value {val}: mean = {mean:.2f}, std dev = {std_dev:.2f}"
      plt.title(title)
      plt.show()


  def normal_distribution_label(self,label, columns):
    ones_label = self.df_model[self.df_model[label] == 1]
    zero_label = self.df_model[self.df_model[label] == 0]
    for i in columns:
      plt.figure(figsize=(10,10))
      ones_label[i].plot(kind="kde", color="green", title=i)
      zero_label[i].plot(kind="kde", color="red")
      plt.xlim([0,max(ones_label[i])])

  def dummies(self, columns):
    df = pd.get_dummies(self.df_model, columns = columns)
    self.df_model = df


  def min_max_scaler(self, columns):
    clf = MinMaxScaler()
    df_min_max = self.df_model[columns]
    data_transformed = clf.fit_transform(df_min_max.to_numpy())
    data_transformed = pd.DataFrame(data_transformed, columns=columns)
    df_one_hot = pd.concat([self.df_model.drop(columns, axis=1), data_transformed], axis=1)
    self.df_model = df_one_hot

  def standard_scaler(self, columns):
    clf = StandardScaler()
    df_min_max = self.df_model[columns]
    data_transformed = clf.fit_transform(df_min_max.to_numpy())
    data_transformed = pd.DataFrame(data_transformed, columns=columns)
    df_one_hot = pd.concat([self.df_model.drop(columns, axis=1), data_transformed], axis=1)
    self.df_model = df_one_hot

  def best_features(self, n_features, steps, data_percentage, label_predict, type_problem, cache_size=5000):
    '''
      Uses RFE  to predict key features for model
      For type_problem = "regression" it uses SVR to extract key features
    '''
    if type_problem == "regression":
      df = self.df_model.sample(frac=data_percentage)
      X = df.drop(label_predict, axis=1)
      y = df[label_predict]
      estimator = SVR(kernel="linear", cache_size=cache_size)
      selector = RFE(estimator, n_features_to_select=n_features, step=steps)
      selector.fit(X,y)
      top_features = X.columns[selector.support_]
      return top_features


from sklearn.model_selection import train_test_split, GridSearchCV, KFold
from sklearn.ensemble import HistGradientBoostingClassifier
from ..metrics import Metrics
import optuna.visualization as vis
from sklearn.metrics import accuracy_score, roc_auc_score
from joblib import dump, load
import numpy as np
import optuna
optuna.logging.disable_default_handler()


class HistGradient(Metrics):
    def __init__(self):
        self.metric = Metrics()
        self.model = None
        self.parameters = None

    def create(self,X,y,params=None):
        if params == None:
            hist = HistGradientBoostingClassifier()
            hist.fit(X,y)
            self.model = hist
        else:
            hist = HistGradientBoostingClassifier(**params)
            hist.fit(X,y)
            self.model = hist
            self.parameters = params

    def create_grid(self, X, y, params=None, cv=3):
        params_columns = ["loss", "learning_rate", "max_iter",
                          "max_leaf_nodes","max_depth","min_samples_leaf"
                          ,"l2_regularization","max_bins","validation_fraction","verbose",
                          "random_state"]
        params_basic = {
            'loss': ['log_loss'],
            'learning_rate': [0.01, 0.1],
            'max_iter': [200, 1000],
            'max_leaf_nodes': [13, 31],
            'max_depth': [3, 15, None],
            'min_samples_leaf': [10, 30],
            'l2_regularization': [0, 2],
            'max_bins': [255],
            'validation_fraction': [0.2],
            'verbose': [0],
            'random_state': [42]
        }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        hist = HistGradientBoostingClassifier()
        grid_search = GridSearchCV(hist, params, cv=cv)
        grid_search.fit(X,y)
        self.model = grid_search.best_estimator_
        self.parameters = grid_search.best_params_

    def create_optuna(self, X, y, params=None, n_trials=2, show_plot=False, show_features=False):
        params_columns = ["loss", "learning_rate", "max_iter",
                            "max_leaf_nodes","max_depth","min_samples_leaf"
                            ,"l2_regularization","max_bins","validation_fraction","verbose",
                            "random_state"]
        params_basic = {
            'loss': ['log_loss'],
            'learning_rate': [0.01, 0.1],
            'max_iter': [200, 1000],
            'max_leaf_nodes': [13, 31],
            'max_depth': [1, 7],
            'min_samples_leaf': [10, 30],
            'l2_regularization': [0, 2],
            'max_bins': 255,
            'validation_fraction': 0.2,
            'verbose': 0,
            'random_state': 42
        }
        if params == None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.2,random_state=42)

        def objective(trial):
            param = {
                'loss': trial.suggest_categorical("loss", params['loss']),
                'learning_rate': trial.suggest_loguniform('learning_rate', params['learning_rate'][0], params['learning_rate'][1]),
                'max_iter': trial.suggest_int('max_iter', params['max_iter'][0], params['max_iter'][1]),
                'max_leaf_nodes': trial.suggest_int('max_leaf_nodes', params['max_leaf_nodes'][0], params['max_leaf_nodes'][1]),
                'max_depth': trial.suggest_int('max_depth', params['max_depth'][0], params['max_depth'][1]),
                'min_samples_leaf': trial.suggest_int('min_samples_leaf', params['min_samples_leaf'][0], params['min_samples_leaf'][1]),
                'l2_regularization': trial.suggest_float('l2_regularization', params["l2_regularization"][0], params["l2_regularization"][1]),
                'max_bins': trial.suggest_int('max_bins', params['max_bins'], params['max_bins']),
                'validation_fraction': trial.suggest_int("validation_fraction", params['validation_fraction'], params['validation_fraction']),
                'verbose': trial.suggest_int("verbose", params['verbose'], params['verbose']),
                'random_state': trial.suggest_int("random_state", params["random_state"], params["random_state"])
            }
            hist = HistGradientBoostingClassifier(**param)
            hist.fit(X_train, y_train)
            preds = hist.predict(X_test)
            accuracy = accuracy_score(y_test, preds)
            return accuracy

        study = optuna.create_study(direction='maximize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = vis.plot_optimization_history(study)
            optimization_history_plot.show()
        if show_features:
            param_importance_plot = vis.plot_param_importances(study)
            param_importance_plot.show()
            
        best_params = study.best_params
        hist_best = HistGradientBoostingClassifier(**best_params)
        hist_best.fit(X, y)
        self.model = hist_best
        self.parameters = best_params


    def score(self, X, y):
        preds = self.model.predict(X)
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        return self.model.predict(X)
    
    def evaluate_kfold(self, X, y, df_test, n_splits=5, params=None):
        if params == None:
            params = self.parameters
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        predictions = np.zeros(df_test.shape[0])
        roc = []
        n=0

        for i, (train_index, valid_index) in enumerate(kfold.split(X,y)):
            X_train, X_test = X.iloc[train_index], X.iloc[valid_index]
            y_train, y_test = y.iloc[train_index], y[valid_index]

            self.create(X_train,y_train,params=params)
            predictions += self.predict(df_test)/n_splits
            val_pred = self.predict(X_test)
            roc.append(roc_auc_score(y_test,val_pred))

            print(f"{i} Fold scored: {roc[i]}")

        print(f"Mean roc score {np.mean(roc)}")
        return predictions

    def get(self):
        return self.model
    
    def get_parameters(self):
        return self.parameters
    
    def save(self, model_path="hist_gradient.joblib"):
        dump(self.model, model_path)
    
    def load(self, model_path):
        self.model = load(model_path)


import tensorflow as tf
from sklearn.metrics import log_loss
import optuna.visualization as vis
from ...classification.metrics import Metrics
from transformers import TFAutoModel
import optuna
import numpy as np
optuna.logging.disable_default_handler()

class TransformerModel(Metrics):
    def __init__(self,seq_len=128) -> None:
        self.input_ids = tf.keras.layers.Input(shape=(seq_len,), name='input_ids', dtype='int64')
        self.mask = tf.keras.layers.Input(shape=(seq_len,), name='attention_mask', dtype='int64')

        self.embedding_layer = None
        self.layers = []
        self.output_layer = None
        self.compile_inst = {}
        self.model = None

    def add_layer(self, dimensions=128, activation=None):
        self.layers.append(tf.keras.layers.Dense(dimensions, activation=activation))

    def embedding(self, model_name):        
        model_extract = TFAutoModel.from_pretrained(model_name)
        
        if 'gpt' in model_name or 'roberta' in model_name:
            outputs = model_extract(self.input_ids)
            embeddings = outputs.last_hidden_state
        else:
            outputs = model_extract(self.input_ids, attention_mask=self.mask)
            embeddings = outputs.last_hidden_state
        self.embedding_layer = embeddings

    def output(self, dimension=1, activation="sigmoid"):
        self.output_layer = tf.keras.layers.Dense(dimension, activation=activation)
    
    def compile(self, loss, optimizer, metrics):
        self.compile_inst = {"loss": loss,
                            "optimizer": optimizer,
                            "metrics":metrics}

    def fit(self, train, val, epochs=3, layer_untrained=0):
        embed = self.embedding_layer
        x = self.layers[0](embed)
        for i in range(1, len(self.layers)):
            x = self.layers[i](x)
        y = self.output_layer(x)
        if self.mask is not None:
            model = tf.keras.Model(inputs=[self.input_ids, self.mask], outputs=y)
        else:
            model = tf.keras.Model(inputs=self.input_ids, outputs=y)
        model.compile(loss=self.compile_inst["loss"],
                           optimizer=self.compile_inst["optimizer"],
                           metrics=self.compile_inst["metrics"])
        model.layers[layer_untrained].trainable = False
        model.fit(train, validation_data=val, epochs=epochs)
        self.model = model

    def fit_optuna_layers(self, train, val, test_data, test_labels, params=None, n_trials=10, show_plot=False):
        params_columns = ["layers","layers_dimensions",]
        params_basic = {
            "layers": [2, 10],
            "layers_dimensions": [32, 1024],
        }
        if params is None:
            params = params_basic
        else:
            for parameter in params_columns:
                if parameter not in params.keys():
                    params[parameter] = params_basic[parameter]

        def objective(trial):
            self.reset_layers()
            layers = trial.suggest_int("num_layers", params["layers"][0], params["layers"][1])
            for i in range(layers):
                self.add_layer(
                    trial.suggest_int(f"Layer_{i}", params["layers_dimensions"][0], params["layers_dimensions"][1]),
                    activation="relu"
                )
            self.fit(train, val, epochs=1)
            print("hereee")
            preds = self.predict(test_data)
            print("but its here")
            preds = np.squeeze(preds, axis=1)
            print("ol no ")
            loss = log_loss(test_labels, preds)
            return loss

        study = optuna.create_study(direction='minimize', pruner=optuna.pruners.MedianPruner())
        study.optimize(objective, n_trials=n_trials)
        if show_plot:
            optimization_history_plot = optuna.visualization.plot_optimization_history(study)
            optimization_history_plot.show()

        best_params = study.best_params
        self.reset_layers()
        for layer in range(best_params["layers"]):
            self.add_layer(best_params[f"Layer_{layer}"], activation="relu")
        self.fit(train, val, epochs=2)
                
    def evaluate(self, X, y):
        preds = self.model.predict(X)
        return self.metric.calculate_metrics(y, preds)

    def predict(self, X):
        predictions = []
        for input_ids, attention_mask in X:
            preds = self.model.predict([input_ids, attention_mask])
            predictions.append(preds)
        predictions = np.concatenate(predictions, axis=0)
        return predictions
    
    def reset_layers(self):
        self.layers = []

    def show_layer(self):
        for layer in self.layers:
            print(f"{layer}")





    


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

VERSION = '1.2.0'
DESCRIPTION = 'Auto_AI_patryk'
LONG_DESCRIPTION = 'Package contains helping functions for creating ML models'

# Read the requirements from requirements.txt
with open('./KaggleAutoAI/requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="PatryksAutoAI",
    version=VERSION,
    author="patryk",
    author_email="lyczkopatryk1@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=requirements, 
    keywords=['python', 'machine_learning', 'auto'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
    ]
)



from ...classification.models.logistic_regression import LogisticRegression
from ...model_data.model_data import ModelData
from ..algorithms.complexity_level0 import Complexity_Level0
from ..algorithms.complexity_level1 import Complexity_Level1
import logging
logging.basicConfig(level=logging.WARNING)

class LogisticRegressionAutomated(ModelData):
    def __init__(self, cat):
        self.model = None
        self.columns_used = None
        
    def create(self, X, y, complexity):
        if complexity == 0:
            lr = LogisticRegression()
            lr.create_optuna(X,y,n_trials=200)
            self.model = lr.get()
            self.columns_used = X.columns
        if complexity == 1:
            lr = LogisticRegression()
            self.model, self.columns_used = Complexity_Level0(lr, X, y)

    def get(self):
        return self.model
    
    def columns(self):
        return self.columns_used